/*****************************************************************
/
/ File   :   ctapi.c
/ Skeleton :   David Corcoran
/ Changes for Intertex modem/reader : Mrten Norman
/ Date   :   Februari, 1999
/ Purpose:   Defines CT-API functions
/ License:   See file LICENSE
/
******************************************************************/

#include "defines.h"
#include "ctapi.h"
#include "ixapi.h"
#include "ixraw.h"

/*
 * Defines for this file
 */
#define TRASHBUFLEN 270  /* Length of a buffer big enough for any return */

/*
 *  Static variables for this file
 */
static unsigned char protocol;  /* Protocol of active card */

/*
 * Translates return value from the Intertex reader into ctapi
 * values.
 */
int ixret2ctret(int ixret)
{
  switch (ixret) {
  case IX_ICC_OK:
  case IX_ICC_UNNORMAL_SW1_2: /* Let application detect SW1+SW2 itself */
  case IX_ICC_SW1_2_TOO_EARLY:
    return OK;
  case IX_ICC_ILLEGAL_CMD:
  case IX_ICC_ILLEGAL_PAR:
  case IX_ICC_BAD_LENGTH:
    return ERR_INVALID;
  case IX_ICC_CRD_REMOVED:
    return ERR_CT;
  case IX_ICC_CRD_PAR_ERRORS:
  case ICC_ERR_OPENFAIL:
  case ICC_ERR_DCBFAIL:
  case ICC_ERR_TIMOSETUPFAIL:
  case ICC_ERR_ATSCFAIL:
  case ICC_ERR_MSGFAIL:
  case ICC_ERR_MSGCORRUPT:
  case ICC_ERR_NOTOPEN:
    return ERR_TRANS;
  case ICC_ERR_BUFTOOSMALL:
    return ERR_MEMORY;
  default: /* If nothing else fits */
    return ERR_CT;
  }
}

/* Initializes the port on which the reader resides */

int  CT_init ( unsigned int Ctn,  unsigned int pn ) {
  
  bool BretVal;        /* Return value from IO_InitializePort() */
  int  IretVal;        /* Return for this function */  

/* CHANGE :: You may have to change some of the IO_InitializePort() 
   parameters such as baudrate, parity, and bits.                   */

  switch( pn ) {

  case PORT_COM1:

    #ifdef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/cua/a");
    #endif

    #ifndef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/ttyS0");
    #endif

    break;

  case PORT_COM2:

    #ifdef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/cua/b");
    #endif

    #ifndef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/ttyS1");
    #endif

    break;

  case PORT_COM3:

    #ifdef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/cua/c");
    #endif

    #ifndef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/ttyS2");
    #endif

    break;

  case PORT_COM4:

    #ifdef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/cua/d");
    #endif

    #ifndef SUN_SPARC
     BretVal = ICC_Open_Channel("/dev/ttyS3");
    #endif

    break;

  case PORT_Printer:

   BretVal = ICC_ERR_OPENFAIL;
   break;

  case PORT_Modem:

   BretVal = ICC_Open_Channel("Modem Port");
   break;

  default:
    BretVal = ICC_Open_Channel("/dev/modem");
    break;
  }
  
  if (BretVal) {
    IretVal = ERR_MEMORY;        /* Could not allocate port */
  } else {
    IretVal = OK;
  }
  
  return IretVal;  
}

/* Closes the port in which the reader resides */

int  CT_close( unsigned int Ctn ) {
  
  if (ICC_Close_Channel() == TRUE) {
    return OK;
    
  } else {
    return ERR_CT;
    
  }
}            

/* 
 * Sends/Receives Data to/from the Reader
 * Parameters:
 *      ctn: logical cardterminal number.
 *	dad: destination address.
 *	sad: source address.
 *	lc: cmd length in bytes.
 *	cmd: ICC cmd or CT cmd.
 *	lr: passing of the max. buffer size of the rsp and 
 *	return of the actual length of the rsp.
 * 	rsp: rsp to the cmd.
 * Returns:
 *      OK or CT-API error code.
 */ 

int  CT_data( unsigned int ctn, unsigned char *dad, unsigned char *sad,
              unsigned int  lc, unsigned char *cmd, unsigned int  *lr,
              unsigned char *rsp ) {
  
  /* Reader specific CT-BCS commands */
  

  int IretVal = ERR_INVALID;     /* Return Value. Assume error.    */
  int IXret;                     /* Return from IX API call */
  int lrtmp;                     /* Saves return buffer length temporary */
  unsigned char buf[TRASHBUFLEN]; /* Trash return buffer */
  unsigned char ixcmd;           /* Command echo return */
  unsigned char ixpar;           /* Parameter return */

  if ( *dad == 1 ) {             /* This command goes to the reader */

    *sad = 1;                   /* Source Reader    */
    *dad = 2;                   /* Destination Host */

  /*******************/
  /* CT-BCS Commands */
  /*******************/

  /* Request ICC  - Turns on the reader/card */

    if ( (cmd[0] = 0x20) && (cmd[1] == 0x12) ) {
      /*
       * Insert code here to flash CRD/IN on display and waiting for card,
       * if/when PC/SC supports it.
       */

      if (*lr < 2) return ERR_MEMORY; /* Return buffer too small */

      /* Try T=0 */
      lrtmp = TRASHBUFLEN;
      protocol = 0;
      IXret = ICC_Activate_Card(protocol, buf, &lrtmp);

      if (IXret == IX_ICC_CRD_IS_T1) { /* It was T=1! */
	lrtmp = TRASHBUFLEN;
	protocol = 1;
	IXret = ICC_Activate_Card(protocol, buf, &lrtmp);
      }

      if (IXret == IX_ICC_WRONG_CRD_TYPE) { /* Strange card */
	/* Insert code to test/activate other than T=0/1 here, if needed. */
      }

      IretVal= ixret2ctret(IXret);
      *lr = 2;
      switch (IXret) {
      case IX_ICC_OK:
	rsp[0] = 0x90; /* Asynchronous ICC presented */
	rsp[1] = 0x01;
	break;
      case IX_ICC_CRD_REMOVED:
	rsp[0] = 0x62; /* Card not present */
	rsp[1] = 0x00;
	break;
      default:
	rsp[0] = 0x64; /* Not successful */
	rsp[1] = 0x00;
	break;
      }


  /* Resets the Card/Terminal and returns Atr */

  } else if ( (cmd[0] = 0x20) && (cmd[1] == 0x11) ) {

    if (*lr < 2) return ERR_MEMORY; /* Return buffer too small */
    *lr -=2; /* Reserve 2 bytes for SW1+SW2 */

    /* Get entire ATR */
    IXret = IX_CrdIO(1, 0, 0, 0 /*NULL*/, &ixcmd, &ixpar, lr, rsp);

    IretVal= ixret2ctret(IXret);
      switch (IXret) {
      case IX_ICC_OK:
	rsp[(*lr)++] = 0x90; /* successful */
	rsp[(*lr)++] = 0x00;
	break;
      default:
	*lr = 2;
	rsp[0] = 0x64; /* Not successful */
	rsp[1] = 0x00;
	break;
      }

  /* Get Status - Gets reader status */

  } else if ( (cmd[0] = 0x20) && (cmd[1] == 0x13) ) {

    if (*lr < 3) return ERR_MEMORY; /* Return buffer too small */
    /* Ask for status */
    lrtmp = 0;
    IXret = IX_CrdIO(3, 0, 0, 0 /*NULL*/, &ixcmd, &ixpar, &lrtmp, 0 /*NULL*/);

    *lr = 3;
    switch (ixpar) {
    case 1: /* Card inserted, but inactive */
    case 2: /* Card there, and active */
      rsp[0] = 4;
      break;
    case 3: /* Card not inserted */
      rsp[0] = 0;
      break;
    case 4: /* Card not inserted, but has visited since last check */
      rsp[0] = 1;
      break;
    default:
      /* Should not happen. Report as error. */
      IXret = IX_ICC_ILLEGAL_PAR;
    }

    IretVal= ixret2ctret(IXret);
    rsp[1] = 0x90; /* successful */
    rsp[2] = 0x00;

  /* Eject ICC - Deactivates Reader  */

  } else if ( (cmd[0] = 0x20) && (cmd[1] == 0x15) ) {

    if (*lr < 2) return ERR_MEMORY; /* Return buffer too small */
    IXret = ICC_Deactivate_Card();
    IretVal= ixret2ctret(IXret);
    *lr = 2;
    rsp[1] = 0x90; /* successful */
    rsp[2] = 0x00;

  } else {

    /* Use ICC_Do_Card_Command() with parameters here, if needed. */
    /* CHANGE:: Write directly to the reader. */
  }

 } else if ( *dad == 0 ) {      /* This command goes to the card */
    
    // Don't get confused here this is for the return saying
    // the source was the card and the destination the host

    *sad = 0;  /* Source Smartcard */
    *dad = 2;  /* Destination Host */
    

    if (protocol == 0) {
      /*
       * Assume send command if more than <CLA><INS><P1><P2><P3> = 5 bytes
       * otherwise, assume receive.
       *
       * Feels a bit risky (might be some I/O-command cards out there??),
       * but this appears to be how the others are doing it...
       */
      if (lc > 5) 
	IXret = ICC_Data_To_Card(cmd, lc, rsp, lr, 0);
      else
	IXret = ICC_Data_From_Card(cmd, lc, rsp, lr);
    }
    else { /* (T=1 should work, for the others: try it and hope...) */
      IXret = ICC_Data_From_Card(cmd, lc, rsp, lr);
    }
    IretVal= ixret2ctret(IXret);

 } else {
   IretVal = ERR_INVALID;              /* Invalid SAD/DAD Address */
 }

  if (IretVal != OK) {
    *lr = 0;
  }
  
  return IretVal;
}
